"""Module for the OdkChoices class."""
from ppp.definitions.error import InvalidLanguageException
from ppp.definitions.constants import CHOICE_NAME_VARIATIONS


class OdkChoices:
    """A class to represent a choice list defined in an XLSForm.

    Attributes:
        list_name (str): The name of the choice list.
        data (list): A list of choice options for the choice list.

    """

    def __init__(self, list_name):
        """Initialize a choice list.

        Args:
            list_name (str): The name of the choice list.
        """
        self.list_name = list_name
        self.data = []

    def __repr__(self):
        """Print representation of instance."""
        return "<Odkchoices '{}(len: {})'>".format(self.list_name,
                                                   len(self.data))

    def __str__(self):
        """Convert instance to string."""
        return '{}: {}'.format(self.list_name, self.labels(lang='English'))

    def add(self, choice):
        """Add a choice to this choice list.

        Args:
            choice (dict): A single choice row.
        """
        self.data.append(choice)

    def labels(self, lang):
        """Get the labels for this choice list in the desired language.

        Args:
            lang (str): The language in which to return the choice labels.

        Returns:
            list: Correctly ordered list of choice labels.

        Raises:
            InvalidLanguageException
        """
        label_variations = ['label']
        if lang:
            label_variations = \
                [x.format(lang) for x in ('label::{}', 'label:{}')]
            label_variations.append('label')
        try:
            for label in label_variations:
                if label in self.data[0]:
                    return [d[label] for d in self.data]
        except (KeyError, IndexError):
            msg = 'Language {} not found in choice list {}.'\
                .format(lang, self.list_name)
            raise InvalidLanguageException(msg)

    def name_labels(self, lang):
        """Get choice name labels.

        Args:
            lang (str): The language of choice list.

        Returns:
            list: Choice variable names and associated labels for choice list.
        """
        rows = enumerate(self.data)
        labels = self.labels(lang)
        formatted_rows = []
        for i, row in rows:
            formatted_row = {'label': labels[i]}
            for x in CHOICE_NAME_VARIATIONS:
                if x in row:
                    formatted_row['name'] = row[x]
            formatted_rows.append(formatted_row)
        return formatted_rows

    def choice_langs(self):
        """Discover all languages for these choices.

        Returns:
            list: Alphabetized list of languages.

        Raises:
            InvalidLanguageException: If languages of first row differ from
                languages in any other row.
        """
        langs = set()
        for datum in self.data:
            these_langs = set()
            for k in datum:
                if k == 'label':
                    these_langs.add('')  # Default language
                elif k.startswith('label::'):
                    lang = k[len('label::'):]
                    these_langs.add(lang)
                elif k.startswith('label:'):
                    lang = k[len('label:'):]
                    these_langs.add(lang)
            if not langs:
                langs = these_langs
            elif langs != these_langs:
                msg = 'In choice list {}, different languages found: {} and {}'
                msg = msg.format(self.list_name, langs, these_langs)
                raise InvalidLanguageException(msg)
        lang_list = sorted(list(langs))
        return lang_list



"""Error classes for package."""


class OdkException(Exception):
    """General ODK exception."""

    pass


class OdkFormError(OdkException):
    """General OdkForm error."""

    pass


class OdkChoicesError(OdkException):
    """General OdkChoices error."""

    pass


class InvalidLanguageException(OdkException):
    """General error related to language of ODK form."""

    pass


class AmbiguousLanguageError(InvalidLanguageException):
    """Ambiguous language error."""

    pass


class InconsistentLabelLanguage(InvalidLanguageException):
    """Inconsistent label language error."""

    pass



"""Module for the OdkGroup class."""
# from ppp.config import TEMPLATE_ENV
from ppp.config import get_template_env
from ppp.odkprompt import OdkPrompt, set_template_env as odkpromt_template
from ppp.odktable import OdkTable, set_template_env as odktable_template
from ppp.definitions.utils import exclusion

TEMPLATE_ENV = None


def set_template_env(template):
    """Set template env."""
    global TEMPLATE_ENV
    TEMPLATE_ENV = get_template_env(template)
    odktable_template(template)
    odkpromt_template(template)


class OdkGroup:
    """Class to represent a field-list group in XLSForm.

    Attributes:
        opener (dict): A dictionary row representing first row of group.
        data (list): A list of group components.
        pending_table (OdkTable): A variable for storing an OdkTable object as
            it is being constructed.
        in_repeat (bool): Is this group part of a repeat group?
    """

    def __init__(self, opener):
        """Initialize a group.

        Args:
            opener (dict): A dictionary row representing first row of group.
                In ODK Specification, this would be of 'begin group' type.
        """
        self.opener = opener
        self.data = []
        self.pending_table = None
        self.in_repeat = False

    def __repr__(self):
        """Print representation."""
        return "<OdkGroup {}: {}>".format(self.opener['name'], self.data)

    @staticmethod
    def format_header(header):
        """Render group header.

        A group header is an OdkPrompt with a few extra attributes. Header is
        formatted and returned before initializating as an OdkPrompt.

        Args:
            header (dict): A dictionary row representing first row of group.

        Returns:
            dict: A reformatted representation.
        """
        header['in_group'] = True
        header['simple_type'] = header['type']
        header['is_group_header'] = True
        return header

    def add(self, row):
        """Add a row of data from XLSForm.

        This method should not be called if row comes from an ODK table.

        Args:
            row (dict): Row from XLSForm.
        """
        self.add_pending()
        self.data.append(row)

    def add_table(self, odkprompt):
        """Add a row of a table object to the group.

        This method should only be called for rows from an ODK table.

        Args:
            odkprompt (OdkPrompt): Row from XLSForm.
        """
        if self.pending_table:
            self.pending_table.add(odkprompt)
        else:
            self.pending_table = OdkTable()
            self.pending_table.add(odkprompt)

    def add_pending(self):
        """Add table to group data if one is in being built."""
        if self.pending_table:
            self.data.append(self.pending_table)
            self.pending_table = None

    def to_text(self, lang):
        """Get the text representation of the full group.

        Args:
            lang (str): The language.

        Returns:
            str: The text for this group.
        """
        obj_texts = (d.to_text(lang) for d in self.data)
        sep = '\n\n{}\n\n'.format(' -' * 25)
        group_text = sep.join(obj_texts)
        return group_text

    def to_html(self, lang, **kwargs):
        """Convert group components to html and return concatenation.

        Args:
            lang (str): The language.

        Returns:
            str: A rendered html concatenation of component templates.
        """
        html = ''
        # pylint: disable=no-member

        # - Render header
        html += TEMPLATE_ENV.get_template('content/group/group-opener.html')\
            .render(**kwargs, settings=kwargs)
        header = self.format_header(self.opener)

        html += OdkPrompt(header).to_html(lang, **kwargs)

        # - Render body
        for i in self.data:
            if exclusion(item=i, settings=kwargs):
                continue

            if isinstance(i, OdkPrompt):
                i.row['in_repeat'] = self.in_repeat
                i.row['in_group'] = True
                html += i.to_html(lang, **kwargs)
            elif isinstance(i, OdkTable):
                i.in_repeat = self.in_repeat
                html += i.to_html(lang, **kwargs)

        # - Render footer
        # pylint: disable=no-member
        html += TEMPLATE_ENV.get_template('content/group/group-closer.html')\
            .render(**kwargs, settings=kwargs)

        return html



"""Configuration settings for PPP package."""
from jinja2 import Environment, PackageLoader
import re


EXCLUSION_TOKEN = 'X'


def question_number(question_num, max_length=4):
    """Splitting question number.

    Splitting string first by separator chars, then separating
    letters from numbers, and then splitting just by length.
    All actions performed one-by-one and only if needed.

    Args:
        question_num: Question number in a string representation.
        max_length: Max length of each chunk, with a default value = 4.
        
    Returns:
        Original question number, splitted to (=< max_lengh) chunks
        and then joined to one string with a space char as a glue.
    """
    # If question's length is less then limit, return immediately.
    if len(question_num) <= max_length:
        return question_num

    # Split question number by separator chars.
    while True:
        pieces = re.split("[^A-Za-z0-9 ]", question_num, 1)
        if len(pieces) == 1:
            break
        elif len(pieces[1]) > max_length:
            question_num = ' '.join(pieces)
            continue
        else:
            break

    # If string chunk is longer then limit, split numbers from letters in it.
    pieces = (' '.join(pieces)).split(' ')
    for index, value in enumerate(pieces):
        if len(value) > max_length:
            pieces[index] = \
                ' '.join(map(str.strip, re.findall('(\d+|\D+)', value)))

    # If string chunk is still longer then limit, split it by limit chars.
    pieces = (' '.join(pieces)).split(' ')
    for index, value in enumerate(pieces):
        if len(value) > max_length:
            pieces[index] = ' '.join(
              [value[i:i+max_length] for i in range(0, len(value), max_length)]
            )

    return ' '.join(pieces)


def get_template_env(template):
    """Get Jinja2 template environment.

        Args:
            template (string): The template chosen.

        Returns:
            jinja2.Environment: The environment of chosen template.
    """
    env = Environment(loader=PackageLoader('ppp', 'templates/' + template),
                      trim_blocks=True,
                      lstrip_blocks=True)
    env.filters['question_number'] = question_number
    return env



"""Module for the OdkRepeat class."""
import textwrap
# from ppp.config import TEMPLATE_ENV
from ppp.config import get_template_env
from ppp.odkgroup import OdkGroup, set_template_env as odkgroup_template
from ppp.odkprompt import OdkPrompt, set_template_env as odkpromt_template
from ppp.odktable import OdkTable, set_template_env as odktable_template
from ppp.definitions.utils import exclusion

TEMPLATE_ENV = None


def set_template_env(template):
    """Set template env"""
    global TEMPLATE_ENV
    TEMPLATE_ENV = get_template_env(template)
    odkgroup_template(template)
    odkpromt_template(template)
    odktable_template(template)


class OdkRepeat:
    """Class to represent repeat construct from XLSForm.

    Attributes:
        opener (dict): A dictionary row representing first row of repeat group.
        data (list): A list of repeat group components.

    """

    def __init__(self, opener):
        """Initialize a repeat group.

        Args:
            opener (dict): A dictionary row representing first row of group.
                In ODK Specification, this would be of 'begin repeat' type.
        """
        self.opener = opener
        self.data = []

    def __repr__(self):
        """Print representation of instance."""
        return "<OdkRepeat {}: {}>".format(self.opener['name'], self.data)

    @staticmethod
    def render_header(i, lang, **kwargs):
        """Render repeat group header.

        A repeat group header consists of some opening html tags, followed by
        an OdkPrompt with a few extra attributes.

        Args:
            i (dict): A dictionary row representing first row of repeat group.
            lang (str): The language.


            **kwargs: Keyword arguments.

        Returns:
            str: A rendered html representation of repeat group header.
        """
        # pylint: disable=no-member
        html = TEMPLATE_ENV.get_template('content/repeat/repeat-opener.html')\
            .render(**kwargs, settings=kwargs)
        i['simple_type'] = i['type']
        i['in_repeat'] = True
        i['is_repeat_header'] = True
        html += OdkPrompt(i).to_html(lang, **kwargs, settings=kwargs)
        return html

    @staticmethod
    def render_footer():
        """Render repeat group footer.

        A repeat group footer consists of some closing html tags.

        Returns:
            str: A rendered html representation of repeat group footer.
        """
        # pylint: disable=no-member
        return TEMPLATE_ENV.get_template('content/repeat/repeat-closer.html')\
            .render()

    def add(self, obj):
        """Add XLSForm object to repeat.

        Only other prompts and groups are allowed to be added.

        Args:
            obj (OdkPrompt or OdkGroup): Item to add.
        """
        self.data.append(obj)

    def to_text(self, lang):
        """Get the text representation of the entire repeat group.

        Args:
            lang (str): The language

        Returns:
            str: The text for this repeat.
        """
        obj_texts = (d.to_text(lang) for d in self.data)
        sep = '\n\n{}\n\n'.format('=' * 50)
        repeat_text = sep.join(obj_texts)
        wrapped = textwrap.indent(repeat_text, '|  ', lambda x: True)
        return wrapped

    def to_html(self, lang, **kwargs):
        """Convert repeat group components to html and return concatenation.

        Args:
            lang (str): The language.


            **kwargs: Keyword arguments.

        Returns:
            str: A rendered html concatenation of component templates.
        """
        html = ''

        # - Render header
        html += self.render_header(self.opener, lang, **kwargs)

        # - Render body
        for i in self.data:
            if exclusion(item=i, settings=kwargs):
                continue

            if isinstance(i, OdkPrompt):
                i.row['in_repeat'] = True
                html += i.to_html(lang, **kwargs)
            elif isinstance(i, OdkGroup):
                i.in_repeat = True
                html += i.to_html(lang, **kwargs)
            elif isinstance(i, OdkTable):
                i.in_repeat = True
                # - Render footer
                html += i.to_html(lang, **kwargs)
        html += self.render_footer()

        return html
